module Core where

import Control.Exception                  (IOException)
import Control.Monad.Catch                (catch)
import Data.Aeson                         (FromJSON, ToJSON)
import qualified Data.Text                as T
import Data.ByteString.Lazy.UTF8          (fromString)
import Data.Pool                          (Pool)
import Database.PostgreSQL.Simple         (Connection, Query)
import Database.PostgreSQL.Simple.FromRow (FromRow, field, fromRow)
import Database.PostgreSQL.Simple.ToField (ToField)
import Database.PostgreSQL.Simple.ToRow   (ToRow, toRow)
import Effectful
import Effectful.Dispatch.Dynamic
import Effectful.Error.Static             (Error, throwError)
import Effectful.FileSystem
import Effectful.Reader.Static
import GHC.Generics                       (Generic)
import Servant hiding                     ((:>), throwError)
import qualified Servant                  as S

--
-- Routes
--
type Root     = Get '[PlainText] T.Text
type UserList = "user" S.:> Get '[JSON] [User]
type UserGet  = "user" S.:> Capture "userId"     Int    S.:> Get  '[JSON] User
type UserPost = "user" S.:> ReqBody '[PlainText] String S.:> PostCreated '[PlainText] NoContent

type AppAPI = Root
      :<|> UserList
      :<|> UserGet
      :<|> UserPost

--
-- Core data
--
data User = User { userId :: Int, userName :: String}
  deriving (Show, Generic)

instance FromRow User where
    fromRow = User <$> field <*> field
instance ToRow User where
    toRow (User uid name) = toRow (uid, name)

instance ToJSON User
instance FromJSON User

--
-- Effects
--
type AppEff = Eff '[Database, Reader AppEnv, FileSystem, Logger, Error ServerError, IOE]

data AppEnv = AppEnv { pool :: Pool Connection }

-- Database
data Database :: Effect where
  DatabaseRead  :: (ToField a, Show a, FromRow b) => (Query, a) -> Database (Eff es) [b]
  DatabaseRead_ :: (FromRow b)                    =>  Query     -> Database (Eff es) [b]
  DatabaseWrite :: (ToField a, Show a)            => (Query, a) -> Database (Eff es) ()

type instance DispatchOf Database = 'Dynamic

type DatabaseEffects es = (Reader AppEnv :> es, Logger :> es, Error ServerError :> es, IOE :> es)

databaseRead  :: (ToField a, Show a, Database :> es, Error ServerError :> es) => (Query, a) -> Eff es [User]
databaseRead = send . DatabaseRead

databaseRead_  :: (Database :> es, Error ServerError :> es) => Query -> Eff es [User]
databaseRead_ = send . DatabaseRead_

databaseWrite :: (ToField a, Show a, Database :> es, Error ServerError :> es) => (Query, a) -> Eff es ()
databaseWrite = send . DatabaseWrite

-- Logger
data Logger :: Effect where
  WriteLog :: String -> Logger (Eff es) ()

type instance DispatchOf Logger = 'Dynamic

writeLog :: (Logger :> es, Error ServerError :> es) => String -> Eff es ()
writeLog = send . WriteLog

runLoggerIO :: (IOE :> es, Error ServerError :> es) => Eff (Logger : es) a -> Eff es a
runLoggerIO = interpret $ \_ -> \case
  WriteLog msg -> log msg
  where
    log :: (IOE :> es, Error ServerError :> es) => String -> Eff es ()
    log msg = liftIOE $ putStrLn msg

-- Utility
liftIOE :: (IOE :> es, Error ServerError :> es) => IO a -> Eff es a
-- Lift IO into Eff and catch IOExceptions
liftIOE m = liftIO m `catch` \(e::IOException) -> do
    -- Log IOExceptions to stdout
    liftIO $ putStrLn $ "ERROR Exception: " ++ show e
    -- Throw a custom Servant ServerError
    throwError $ ServerError
      { errHTTPCode     = 500
      , errReasonPhrase = "Internal Server Error"
      , errBody         = "This incident will be investigated."
      , errHeaders      = []
      }
